@@ -14,15 +14,22 @@ from functools import wraps
 
 # Py3k compat.
 if sys.version_info[0] == 3:
-    basestring = str
+    binary_types = (bytes, bytearray)
+    decode_handler = 'backslashreplace'
+    numeric = (int, float)
+    unicode_type = str
     from io import StringIO
 else:
+    binary_types = (buffer, bytes, bytearray)
+    decode_handler = 'replace'
+    numeric = (int, long, float)
+    unicode_type = unicode
     from StringIO import StringIO
 
 try:
     from flask import (
-        Flask, abort, escape, flash, make_response, Markup, redirect,
-        render_template, request, url_for)
+        Flask, abort, escape, flash, jsonify, make_response, Markup, redirect,
+        render_template, request, session, url_for)
 except ImportError:
     raise RuntimeError('Unable to import flask module. Install by running '
                        'pip install flask')
@@ -309,7 +316,6 @@ def add_index(table):
     unique = bool(request_data.get('unique'))
 
     columns = dataset.get_columns(table)
-    column_names = [column.name for column in columns]
 
     if request.method == 'POST':
         if indexed_columns:
@@ -379,20 +385,19 @@ def drop_trigger(table):
 @app.route('/<table>/content/')
 @require_table
 def table_content(table):
-    page_number = int(request.args.get('page') or 1)
-    if page_number > 1:
-        previous_page = page_number - 1
-    else:
-        previous_page = None
+    page_number = request.args.get('page') or ''
+    page_number = int(page_number) if page_number.isdigit() else 1
 
     ds_table = dataset[table]
     total_rows = ds_table.all().count()
     rows_per_page = app.config['ROWS_PER_PAGE']
-    total_pages = math.ceil(total_rows / float(rows_per_page))
-    if page_number < total_pages:
-        next_page = page_number + 1
-    else:
-        next_page = None
+    total_pages = int(math.ceil(total_rows / float(rows_per_page)))
+    # Restrict bounds.
+    page_number = min(page_number, total_pages)
+    page_number = max(page_number, 1)
+
+    previous_page = page_number - 1 if page_number > 1 else None
+    next_page = page_number + 1 if page_number < total_pages else None
 
     query = ds_table.all().paginate(page_number, rows_per_page)
 
@@ -404,8 +409,16 @@ def table_content(table):
         query = query.order_by(field)
 
     field_names = ds_table.columns
-    columns = [field.db_column
-               for field in ds_table.model_class._meta.get_fields()]
+    model_meta = ds_table.model_class._meta
+    try:
+        fields = model_meta.sorted_fields
+    except AttributeError:
+        fields = model_meta.get_fields()
+    columns = [field.db_column for field in fields]
+
+    table_sql = dataset.query(
+        'SELECT sql FROM sqlite_master WHERE tbl_name = ? AND type = ?',
+        [table, 'table']).fetchone()[0]
 
     return render_template(
         'table_content.html',
@@ -418,6 +431,7 @@ def table_content(table):
         previous_page=previous_page,
         query=query,
         table=table,
+        total_pages=total_pages,
         total_rows=total_rows)
 
 @app.route('/<table>/query/', methods=['GET', 'POST'])
@@ -447,6 +461,10 @@ def table_query(table):
         else:
             sql = 'SELECT *\nFROM "%s"' % (table)
 
+    table_sql = dataset.query(
+        'SELECT sql FROM sqlite_master WHERE tbl_name = ? AND type = ?',
+        [table, 'table']).fetchone()[0]
+
     return render_template(
         'table_query.html',
         data=data,
@@ -455,7 +473,18 @@ def table_query(table):
         query_images=get_query_images(),
         row_count=row_count,
         sql=sql,
-        table=table)
+        table=table,
+        table_sql=table_sql)
+
+@app.route('/table-definition/', methods=['POST'])
+def set_table_definition_preference():
+    key = 'show'
+    show = False
+    if request.form.get(key):
+        session[key] = show = True
+    elif key in request.session:
+        del request.session[key]
+    return jsonify({key: show})
 
 def export(table, sql, export_format):
     model_class = dataset[table].model_class
@@ -472,7 +501,7 @@ def export(table, sql, export_format):
 
     dataset.freeze(query, export_format, file_obj=buf, **kwargs)
 
-    response_data= buf.getvalue()
+    response_data = buf.getvalue()
     response = make_response(response_data)
     response.headers['Content-Length'] = len(response_data)
     response.headers['Content-Type'] = mimetype
@@ -536,8 +565,15 @@ def drop_table(table):
 
 @app.template_filter('value_filter')
 def value_filter(value, max_length=50):
-    value = escape(value)
-    if isinstance(value, basestring):
+    if isinstance(value, numeric):
+        return value
+
+    if isinstance(value, binary_types):
+        if not isinstance(value, (bytes, bytearray)):
+            value = bytes(value)  # Handle `buffer` type.
+        value = value.decode('utf-8', decode_handler)
+    if isinstance(value, unicode_type):
+        value = escape(value)
         if len(value) > max_length:
             return ('<span class="truncated">%s</span> '
                     '<span class="full" style="display:none;">%s</span>'
@@ -546,9 +582,10 @@ def value_filter(value, max_length=50):
                         value)
     return value
 
+column_re = re.compile('(.+?)\((.+)\)', re.S)
+column_split_re = re.compile(r'(?:[^,(]|\([^)]*\))+')
+
 def _format_create_table(sql):
-    column_re = re.compile('(.+?)\((.+)\)')
-    column_split_re = re.compile(r'(?:[^,(]|\([^)]*\))+')
     create_table, column_list = column_re.search(sql).groups()
     columns = ['  %s' % column.strip()
                for column in column_split_re.findall(column_list)
@@ -576,7 +613,7 @@ def get_query_images():
     for filename in sorted(os.listdir(image_dir)):
         basename = os.path.splitext(os.path.basename(filename))[0]
         parts = basename.split('-')
-        accum.append((parts, os.path.join('img', filename)))
+        accum.append((parts, 'img/' + filename))
     return accum
 
 #
@@ -659,6 +696,7 @@ def main():
     global migrator
     dataset = SqliteDataSet('sqlite:///%s' % db_file)
     migrator = dataset._migrator
+    dataset.close()
     if options.browser:
         open_browser_tab(options.host, options.port)
     app.run(host=options.host, port=options.port, debug=options.debug)
